package softmc;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javassist.ClassPool;
import javassist.CtClass;
import javassist.NotFoundException;
import javassist.bytecode.ClassFile;
import javassist.bytecode.CodeAttribute;
import javassist.bytecode.ConstPool;
import javassist.bytecode.Descriptor;
import javassist.bytecode.LocalVariableAttribute;
import javassist.bytecode.MethodInfo;
import korat.instrumentation.bytecode.BytecodeInstruction;
import korat.instrumentation.cfg.ast.BasicBlock;
import korat.instrumentation.cfg.ast.CFG;
import korat.instrumentation.cfg.ast.CFGMethodInfo;
import softmc.logic.Factory;
import softmc.logic.Var;
import softmc.state.BlockState;
import softmc.state.GlobalState;
import softmc.state.SymStack;

public class CFG2StatesConverter {

    private final ClassPool cp;
    private ConstPool constPool;
    private CtClass clz;
    private ClassFile classFile;
    private MethodInfo mi; 
    
    public CFG2StatesConverter(ClassPool cp) {
        this.cp = cp;
    } 

    public GlobalState convertToStates(CFG cfg) throws Exception {
        clz = cp.get(cfg.getSrcMethod().className);
        classFile = clz.getClassFile2();
        constPool = classFile.getConstPool();
        mi = findMethod(classFile, cfg.getSrcMethod());
        GlobalState globalState = new GlobalState();
        populateAllVars(globalState, mi);
        Map<BasicBlock, BlockState> visited = new HashMap<BasicBlock, BlockState>();
        SymStack stack = new SymStack();
        globalState.setEntryBlockState(visitBasicBlock(cfg.getEntryBlock(), visited, stack));
        globalState.computeMutableVars();
        return globalState;
    }
    
    static String getLocalVarName(LocalVariableAttribute lva, int idx) {
        return lva.variableName(idx) + idx;
//        if (lva != null) {
//            return lva.variableName(idx);
//        } else {
//            return "local_" + idx;
//        }
    }

    private void populateAllVars(GlobalState globalState, MethodInfo mi) throws NotFoundException {
        Factory factory = Factory.instance;
        CodeAttribute ca = mi.getCodeAttribute();
        LocalVariableAttribute lva = (LocalVariableAttribute) ca.getAttribute("LocalVariableTable");
        int argsNum = Descriptor.getParameterTypes(mi.getDescriptor(), cp).length;
        for (int idx = 0; idx < ca.getMaxLocals(); idx++) {
            String varName = getLocalVarName(lva, idx);
            if ("this".equals(varName))
                continue;
            String varTypeName = Descriptor.toClassName(lva.signature(idx));
            Var var = factory.mkVar(varName, varTypeName);
            globalState.addVar(var);
            if (argsNum-- > 0)
                globalState.addInputParam(var);
        }
    }

    private BlockState visitBasicBlock(BasicBlock bb, Map<BasicBlock, BlockState> visited, SymStack stack) {
        BlockState bbState = visited.get(bb);
        if (bbState != null)
            return bbState;
        bbState = new BlockState();
        visited.put(bb, bbState);
        for (BytecodeInstruction instr : bb.getInstructions()) {
            ConvVisitor visitor = new ConvVisitor(bbState, stack, cp, constPool, mi);
            instr.accept(visitor);
        }
        if (!bbState.isError()) { // if it's ERROR state, don't put outgoing edges
            BasicBlock thenBlock = bb.getThenBlock();
            BasicBlock elseBlock = bb.getElseBlock();
            BlockState thenState = null;
            BlockState elseState = null;
            if (thenBlock != null)
                thenState = visitBasicBlock(thenBlock, visited, stack);
            if (elseBlock != null)
                elseState = visitBasicBlock(elseBlock, visited, stack);
            bbState.update2wayBranchDests(thenState, elseState);
        }
        return bbState;
    }

    @SuppressWarnings("unchecked")
    static MethodInfo findMethod(ClassFile classFile, CFGMethodInfo srcMethod) {
        List<MethodInfo> methods = classFile.getMethods();
        for (MethodInfo mi : methods) {
            if (mi.getName().equals(srcMethod.methodName) && mi.getDescriptor().equals(srcMethod.methodType))
                return mi;
        }
        return null;
    }
    
}
