package softmc;

import static korat.instrumentation.bytecode.BytecodeUtils.GETSTATIC;
import static korat.instrumentation.bytecode.BytecodeUtils.IFNE;

import java.util.HashSet;
import java.util.LinkedList;
import java.util.Set;

import javassist.ClassPool;
import javassist.CtClass;
import javassist.bytecode.ConstPool;
import korat.instrumentation.bytecode.BytecodeInstruction;
import korat.instrumentation.bytecode.BytecodeUtils;
import korat.instrumentation.cfg.ast.BasicBlock;
import korat.instrumentation.cfg.ast.CFG;

public class CFGPreProcessor {

    private final ClassPool cp;
    private ConstPool constPool;
    private CtClass clz;
    
    public CFGPreProcessor(ClassPool cp) {
        this.cp = cp;
    }

    /**
     * Applies all transformations to the given CFG object (by <b>mutating</b> it)
     */
    public void process(CFG cfg) throws Exception {
        clz = cp.get(cfg.getSrcMethod().className);
        constPool = clz.getClassFile2().getConstPool();
        removeAssertionEnabledChecks(cfg);
    } 
    
    /**
     * For this analysis we'll assume that assertions are enabled for the class
     * under test, so we'll simplify the CFG by removing all runtime checks that  
     * check whether assertions are enabled.   
     */
    private void removeAssertionEnabledChecks(CFG cfg) {
        Set<BasicBlock> visited = new HashSet<BasicBlock>();
        LinkedList<BasicBlock> workList = new LinkedList<BasicBlock>();
        workList.add(cfg.getEntryBlock());
        while (!workList.isEmpty()) {
            BasicBlock bb = workList.removeFirst();
            if (!visited.add(bb))
                continue;
            BasicBlock thenBlock = bb.getThenBlock();
            BasicBlock elseBlock = bb.getElseBlock();
            if (thenBlock != null) 
                workList.add(thenBlock);
            if (elseBlock != null)
                workList.add(elseBlock);
            if (thenBlock == null || elseBlock == null) 
                continue; // this is not the "assertion enabled check" block
            int nInstr = bb.getInstructions().size();
            if (nInstr < 2)
                continue;
            BytecodeInstruction instr1 = bb.getInstructions().get(nInstr - 1);
            BytecodeInstruction instr2 = bb.getInstructions().get(nInstr - 2);
            if (instr1.getOpcode() == IFNE && instr2.getOpcode() == GETSTATIC &&
                    BytecodeUtils.GETSTATIC_fieldName(instr2, constPool).equals("$assertionsDisabled")) {
                // that's it
                bb.disconnectThen();
                bb.removeInstruction(nInstr - 1); 
                bb.removeInstruction(nInstr - 2);
                // if the basic block is empty just remove it from the CFG
                if (bb.getInstructions().isEmpty())
                    bb.removeMyself();
            }
        }
        assert cfg.checkRep();
    }

}
