package softmc;

import javassist.ClassPool;
import javassist.CtClass;
import javassist.NotFoundException;
import javassist.bytecode.CodeAttribute;
import javassist.bytecode.ConstPool;
import javassist.bytecode.Descriptor;
import javassist.bytecode.LocalVariableAttribute;
import javassist.bytecode.MethodInfo;
import korat.instrumentation.bytecode.BytecodeInstruction;
import korat.instrumentation.bytecode.BytecodeUtils;
import korat.instrumentation.bytecode.VisitorUnsupported;
import softmc.logic.Factory;
import softmc.logic.IntConst;
import softmc.logic.Sort;
import softmc.logic.Term;
import softmc.logic.Var;
import softmc.logic.visitors.TermTypeInference;
import softmc.logic.visitors.VarExtractor;
import softmc.state.BlockState;
import softmc.state.SymStack;
import softmc.state.SymStack.Entry;

public class ConvVisitor extends VisitorUnsupported {

    private final BlockState bbState; 
    private final SymStack stack;
    private final ConstPool constPool;
    private final Factory factory;
    private final CodeAttribute ca; 
    private LocalVariableAttribute lva;
    private ClassPool cp; 
    
    public ConvVisitor(BlockState bbState, SymStack stack, ClassPool cp, ConstPool constPool, MethodInfo mi) {
        this.bbState = bbState;
        this.stack = stack;
        this.cp = cp;
        this.constPool = constPool; 
        this.ca = mi.getCodeAttribute();
        this.lva = (LocalVariableAttribute) ca.getAttribute("LocalVariableTable"); 
        this.factory = Factory.instance;
    }
    
    private void pushInt(int x) {
        IntConst mkInt = factory.mkInt(x);
        stack.push(mkInt, mkInt);
    }
    
    private void pushVar(Var var) {
        stack.push(var, getVarSymValue(var));
    }
    
    private Term getVarSymValue(Var var) {
        Term symVal = bbState.getVarSymVal(var.getVarName());
        if (symVal == null) {
            symVal = var; 
        }
        return symVal;
    }
    
    public void postVisit(BytecodeInstruction instr) {}
    public void preVisit(BytecodeInstruction instr) {}

    public void visitNOP (BytecodeInstruction instr) {}
    public void visitICONST_M1 (BytecodeInstruction instr) { pushInt(-1); }
    public void visitICONST_0 (BytecodeInstruction instr)  { pushInt(0); }
    public void visitICONST_1 (BytecodeInstruction instr)  { pushInt(1); }
    public void visitICONST_2 (BytecodeInstruction instr)  { pushInt(2); }
    public void visitICONST_3 (BytecodeInstruction instr)  { pushInt(3); }
    public void visitICONST_4 (BytecodeInstruction instr)  { pushInt(4); }
    public void visitICONST_5 (BytecodeInstruction instr)  { pushInt(5); }
    public void visitBIPUSH (BytecodeInstruction instr) { pushInt(instr.byteAt(1)); }
    public void visitSIPUSH (BytecodeInstruction instr) { pushInt(instr.s16bitAt(1)); }
    
    public void visitILOAD (BytecodeInstruction instr)   { visitILoad(instr, instr.byteAt(1)); }
    public void visitILOAD_0 (BytecodeInstruction instr) { visitILoad(instr, 0); }
    public void visitILOAD_1 (BytecodeInstruction instr) { visitILoad(instr, 1); }
    public void visitILOAD_2 (BytecodeInstruction instr) { visitILoad(instr, 2); }
    public void visitILOAD_3 (BytecodeInstruction instr) { visitILoad(instr, 3); }
    
    protected void visitILoad(BytecodeInstruction instr, int idx) {
        String varName = CFG2StatesConverter.getLocalVarName(lva, idx);
        pushVar(factory.mkVar(varName, factory.INT_SORT));
    }
    
    public void visitALOAD (BytecodeInstruction instr)   { visitALoad(instr, instr.byteAt(1)); }
    public void visitALOAD_0 (BytecodeInstruction instr) { visitALoad(instr, 0); }
    public void visitALOAD_1 (BytecodeInstruction instr) { visitALoad(instr, 1); }
    public void visitALOAD_2 (BytecodeInstruction instr) { visitALoad(instr, 2); }
    
    protected void visitALoad(BytecodeInstruction instr, int idx) {
        String varName = CFG2StatesConverter.getLocalVarName(lva, idx);
        pushVar(factory.mkVar(varName, factory.INT_ARRAY_SORT));
    }
    
    public void visitISTORE (BytecodeInstruction instr)   { visitIStore(instr, instr.byteAt(1)); }
    public void visitISTORE_0 (BytecodeInstruction instr) { visitIStore(instr, 0); }
    public void visitISTORE_1 (BytecodeInstruction instr) { visitIStore(instr, 1); }
    public void visitISTORE_2 (BytecodeInstruction instr) { visitIStore(instr, 2); }
    public void visitISTORE_3 (BytecodeInstruction instr) { visitIStore(instr, 3); }
    
    protected void visitIStore(BytecodeInstruction instr, int idx) {
        Entry e = stack.pop();
        String varName = CFG2StatesConverter.getLocalVarName(lva, idx);
        bbState.setVarSymVal(varName, e.expr);
    }
    
    public void visitIALOAD (BytecodeInstruction instr) {
        Entry index = stack.pop();
        Entry array = stack.pop();
        stack.push(factory.mkArraySelect(array.sym, index.sym), 
                factory.mkArraySelect(array.expr, index.expr));
    }
    
    public void visitIASTORE (BytecodeInstruction instr) {
        Entry value = stack.pop();
        Entry index = stack.pop();
        Entry array = stack.pop();
        Var targetVar = new VarExtractor().extractVar(array.expr, factory.INT_ARRAY_SORT);
        bbState.setVarSymVal(targetVar.getVarName(), factory.mkArrayStore(array.expr, index.expr, value.expr));
    }
    
    public void visitPOP (BytecodeInstruction instr)  { stack.pop(); }
    public void visitPOP2 (BytecodeInstruction instr) { stack.pop(); stack.pop(); }
    public void visitDUP (BytecodeInstruction instr)  { stack.push( stack.peek()); }
    public void visitDUP_X1 (BytecodeInstruction instr) { stack.insertAt(2, stack.peek()); }
    //public void visitDUP_X2 (BytecodeInstruction instr) {}
    public void visitDUP2 (BytecodeInstruction instr) { stack.push(stack.peekAt(1)); stack.push(stack.peekAt(1)); }
    //public void visitDUP2_X1 (BytecodeInstruction instr) {}
    //public void visitDUP2_X2 (BytecodeInstruction instr) {}
    public void visitSWAP (BytecodeInstruction instr) { stack.insertAt(1, stack.pop()); }
    public void visitIADD (BytecodeInstruction instr) { 
        Entry rhs = stack.pop();
        Entry lhs = stack.pop();
        stack.push(factory.mkAdd(lhs.sym, rhs.sym), factory.mkAdd(lhs.expr, rhs.expr));
    }
    public void visitISUB (BytecodeInstruction instr) {
        Entry rhs = stack.pop();
        Entry lhs = stack.pop();
        stack.push(factory.mkSub(lhs.sym, rhs.sym), factory.mkSub(lhs.expr, rhs.expr));
    }
    public void visitIMUL (BytecodeInstruction instr) {
        Entry rhs = stack.pop();
        Entry lhs = stack.pop();
        stack.push(factory.mkMul(lhs.sym, rhs.sym), factory.mkMul(lhs.expr, rhs.expr));
    }
    public void visitIDIV (BytecodeInstruction instr) {
        Entry rhs = stack.pop();
        Entry lhs = stack.pop();
        stack.push(factory.mkDiv(lhs.sym, rhs.sym), factory.mkDiv(lhs.expr, rhs.expr));
    }
    public void visitIREM (BytecodeInstruction instr) {
        Entry rhs = stack.pop();
        Entry lhs = stack.pop();
        stack.push(factory.mkMod(lhs.sym, rhs.sym), factory.mkMod(lhs.expr, rhs.expr));
    }
    public void visitINEG (BytecodeInstruction instr) {
        Entry term = stack.pop();
        stack.push(factory.mkSub(factory.mkInt(0), term.sym), 
                factory.mkSub(factory.mkInt(0), term.expr));
    }
    public void visitIINC (BytecodeInstruction instr) {
        int idx = instr.byteAt(1);
        String varName = CFG2StatesConverter.getLocalVarName(lva, idx);
        int incConst = instr.signedByteAt(2);
        Var var = factory.mkVar(varName, factory.INT_SORT);
        bbState.setVarSymVal(varName, factory.mkAdd(getVarSymValue(var), factory.mkInt(incConst)));
    }
    public void visitIFEQ (BytecodeInstruction instr) { 
        Term lhs = stack.pop().sym;
        Sort type = new TermTypeInference().inferType(lhs);
        if (type == factory.BOOL_SORT)
            set2wayBranchCond(factory.mkEq(lhs, factory.mkBool(false)));
        else
            set2wayBranchCond(factory.mkEq(lhs, factory.mkInt(0)));
    }
    public void visitIFNE (BytecodeInstruction instr) { 
        Term lhs = stack.pop().sym;
        Sort type = new TermTypeInference().inferType(lhs);
        if (type == factory.BOOL_SORT)
            set2wayBranchCond(factory.mkEq(lhs, factory.mkBool(true)));
        else
            set2wayBranchCond(factory.mkNeq(lhs, factory.mkInt(0))); 
    }
    public void visitIFLT (BytecodeInstruction instr) { set2wayBranchCond(factory.mkLt(stack.pop().sym, factory.mkInt(0))); }
    public void visitIFLE (BytecodeInstruction instr) { set2wayBranchCond(factory.mkLe(stack.pop().sym, factory.mkInt(0))); }
    public void visitIFGT (BytecodeInstruction instr) { set2wayBranchCond(factory.mkGt(stack.pop().sym, factory.mkInt(0))); }
    public void visitIFGE (BytecodeInstruction instr) { set2wayBranchCond(factory.mkGe(stack.pop().sym, factory.mkInt(0))); }
    public void visitIF_ICMPEQ (BytecodeInstruction instr) { 
        Term rhs = stack.pop().sym; 
        Term lhs = stack.pop().sym; 
        set2wayBranchCond(factory.mkEq(lhs, rhs));
    }
    public void visitIF_ICMPNE (BytecodeInstruction instr) { 
        Term rhs = stack.pop().sym; 
        Term lhs = stack.pop().sym; 
        set2wayBranchCond(factory.mkNeq(lhs, rhs));
    }
    public void visitIF_ICMPLT (BytecodeInstruction instr) { 
        Term rhs = stack.pop().sym; 
        Term lhs = stack.pop().sym; 
        set2wayBranchCond(factory.mkLt(lhs, rhs));
    }
    public void visitIF_ICMPLE (BytecodeInstruction instr) { 
        Term rhs = stack.pop().sym; 
        Term lhs = stack.pop().sym; 
        set2wayBranchCond(factory.mkLe(lhs, rhs));
    }
    public void visitIF_ICMPGT (BytecodeInstruction instr) { 
        Term rhs = stack.pop().sym; 
        Term lhs = stack.pop().sym; 
        set2wayBranchCond(factory.mkGt(lhs, rhs));
    }
    public void visitIF_ICMPGE (BytecodeInstruction instr) { 
        Term rhs = stack.pop().sym; 
        Term lhs = stack.pop().sym; 
        set2wayBranchCond(factory.mkGe(lhs, rhs));
    }
    public void visitGOTO (BytecodeInstruction instr) { set2wayBranchCond(factory.mkBool(true)); }
    public void visitJSR (BytecodeInstruction instr) {}
    public void visitRET (BytecodeInstruction instr) {}
    public void visitIRETURN (BytecodeInstruction instr) { stack.pop(); }
    public void visitLRETURN (BytecodeInstruction instr) { stack.pop(); }
    public void visitFRETURN (BytecodeInstruction instr) { stack.pop(); }
    public void visitDRETURN (BytecodeInstruction instr) { stack.pop(); }
    public void visitARETURN (BytecodeInstruction instr) { stack.pop(); }
    public void visitRETURN (BytecodeInstruction instr)  {}

    public void visitINVOKEVIRTUAL (BytecodeInstruction instr)   {
        invokeInstr(false, false, instr.u16bitAt(1));
    }
    public void visitINVOKESPECIAL (BytecodeInstruction instr)   { 
        invokeInstr(false, false, instr.u16bitAt(1));
    }
    public void visitINVOKESTATIC (BytecodeInstruction instr)    { 
        invokeInstr(false, true, instr.u16bitAt(1));
    }
    public void visitINVOKEINTERFACE (BytecodeInstruction instr) { 
        invokeInstr(true, false, instr.u16bitAt(1));
    }
    
    /**
     * Pops the right number of values from the top of the operand stack (which is
     * equal to the number of method parameters of the method to be invoked), then
     * if the call is not static pops object reference from the operand stack, and
     * finally pushes the return type of the method to the stack.  This method is
     * used for handling various invoke bytecode instructions.
     *
     * @param isInterface - is it an interface method call
     * @param isStatic - is it a static method call
     * @param cpIdx - index in the constant pool for the method to be invoked.
     */
    protected void invokeInstr(boolean isInterface, boolean isStatic, int cpIdx) {
        try {
            String methodDesc;
            if (isInterface) {
                methodDesc = constPool.getInterfaceMethodrefType(cpIdx);
            } else {
                methodDesc = constPool.getMethodrefType(cpIdx);
            }
            CtClass[] params = Descriptor.getParameterTypes(methodDesc, cp);
            Term[] symArgs = new Term[params.length];
            Term[] exprArgs = new Term[params.length];
            for (int i = params.length - 1; i >= 0; i--) {
                Entry e = stack.pop();
                symArgs[i] = e.sym;
                exprArgs[i] = e.expr;
            }
            if (!isStatic) {
                stack.pop(); //objref
            }
            CtClass returnType = Descriptor.getReturnType(methodDesc, cp);
            if (returnType != CtClass.voidType) {
                if (isStatic && !isInterface) {
                    String className = constPool.getMethodrefClassName(cpIdx);
                    if (!Axioms.class.getName().equals(className))
                        throw new RuntimeException("No can do: invoke method of class " + className);
                    String methodName = constPool.getMethodrefName(cpIdx);
                    String retSortName = returnType.getName();
                    stack.push(factory.mkApp(methodName, retSortName, symArgs), 
                            factory.mkApp(methodName, retSortName, exprArgs));
                } else {
                    pushVar(factory.mkVar("$ret", returnType.getName()));
                }
            }
        } catch (NotFoundException e) {
            throw new RuntimeException("Unknown fieldType", e);
        }
    }
    
    public void visitNEW (BytecodeInstruction instr) {
        int cpIdx = instr.u16bitAt(1);
        String className = constPool.getClassInfo(cpIdx);
        pushVar(factory.mkVar("$new_" + className, className));
    }
    
    public void visitARRAYLENGTH (BytecodeInstruction instr) {
        
    }
    
    public void visitNEWARRAY(BytecodeInstruction instr) {
        if (instr.byteAt(1) != BytecodeUtils.T_INT)
            throw new RuntimeException("Only int arrays are supported");
        stack.pop(); // count
        pushVar(factory.mkVar("$newarray", factory.INT_ARRAY_SORT));
    }
    
    public void visitASTORE (BytecodeInstruction instr)   { visitAStore(instr, instr.byteAt(1)); }
    public void visitASTORE_0 (BytecodeInstruction instr) { visitAStore(instr, 0); }
    public void visitASTORE_1 (BytecodeInstruction instr) { visitAStore(instr, 1); }
    public void visitASTORE_2 (BytecodeInstruction instr) { visitAStore(instr, 2); }
    public void visitASTORE_3 (BytecodeInstruction instr) { visitAStore(instr, 3); }
    
    protected void visitAStore(BytecodeInstruction instr, int idx) {
        stack.pop(); // it has to be array because we don't support other referencess
        //String varName = CFG2StatesConverter.getLocalVarName(lva, idx);
        //bbState.setVarSymVal(varName, e.expr);
    }
    
    public void visitATHROW (BytecodeInstruction instr) {
        stack.pop();
        pushVar(factory.mkVar("$throw", "???"));
        bbState.setError(true);
    }
    
    public void visitGOTO_W (BytecodeInstruction instr) { set2wayBranchCond(factory.mkBool(true)); }
    public void visitJSR_W (BytecodeInstruction instr) {}
    
    private void set2wayBranchCond(Term cond) {
        bbState.set2wayBranch(cond, null, null);
    }

}
