package softmc;

import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import korat.utils.collection.CollectionUtils;
import korat.utils.func.Func1;
import korat.utils.func.Pred;
import softmc.state.BlockState;
import softmc.state.GlobalState;
import softmc.state.StateTransition;

public class StateOptimizer {

    /**
     * Removes empty states (those states whose symbolic states are empty) from 
     * the state graph. 
     * 
     * @requires: transitions must not have side effects ("effects" map must be empty)
     */
    public void removeEmptyStates(GlobalState state) {
        BlockState startState = state.getEntryBlockState();
        LinkedList<BlockState> workList = new LinkedList<BlockState>();
        Set<BlockState> visited = new HashSet<BlockState>();
        if (!startState.getSymState().isEmpty())
            workList.push(startState);
        else 
            workList.addAll(startState.getImmSuccStates());
        while (!workList.isEmpty()) {
            BlockState bs = workList.pop();
            if (!visited.add(bs))
                continue;
            List<StateTransition> transitions = bs.getTransitions();
            for (int i = 0; i < transitions.size(); i++) {
                StateTransition tr = transitions.get(i);
                BlockState succ = tr.getDest();
                List<StateTransition> succSuccs = succ.getTransitions();
                if (succSuccs.isEmpty() || !succ.getSymState().isEmpty())
                    continue;   
                boolean b = bs.splitTransition(i, succSuccs);
                if (b) i--;
            }
            workList.addAll(CollectionUtils.select(transitions, new Func1<StateTransition, BlockState>() {
                @Override
                public BlockState exe(StateTransition arg) {
                    return arg.getDest();
                }
            }));
        }
    }
    
    /**
     * Minimizes the state graph by eliminating states that don't contain self-loops.
     */
    public void minimize(GlobalState state) {
        minimize(state, GlobalOptions.__elim_strategy);
    }
    
    private void minimize(GlobalState state, Func1<GlobalState, Collection<BlockState>> strategy) {
        while (true) {
            Collection<BlockState> states = strategy.exe(state);
            if (states.isEmpty()) {
                break;
            }
            BlockState stateToEliminate = states.iterator().next();
            state.eliminateState(stateToEliminate);
        }
    }

    public void unfold(GlobalState state) {
        for (StateTransition trans : CollectionUtils.where(state.getAllTransitions(), new Pred<StateTransition>() {
            @Override public Boolean exe(StateTransition e) {
                return e.getSource() == e.getDest();
            }
        })) {
            trans.getSource().unfoldTransition(trans);
        }
    }

    /**
     * Performs some simple logic transformations on all terms in order to simplify them. 
     */
    public void simplifyTerms(GlobalState state) {
        for (BlockState bs : state.getEntryBlockState().getAllSuccState()) {
            bs.simplifyAllTerms();
        }
    }
    
    // ============================================================================
    // ------------------- StateElimRevtrategy class -----------------------------
    // ============================================================================
    /**
     * Just reverses the order of the list returned by StateElimStrategy
     */
    public static class StateElimRevStrategy implements Func1<GlobalState, Collection<BlockState>> {
        @Override
        public Collection<BlockState> exe(GlobalState globalState) {
            StateElimStrategy str = new StateElimStrategy();
            List<BlockState> l = new LinkedList<BlockState>(str.exe(globalState));
            Collections.reverse(l);
            return l;
        }
    }
    
    // ============================================================================
    // ------------------- StateElimStrategy class --------------------------------
    // ============================================================================
    
    /**
     * Returns all non-leaf states, that contain no self-loops, traversed in 
     * DFS order. 
     */
    public static class StateElimStrategy implements Func1<GlobalState, Collection<BlockState>> {
        @Override
        public Collection<BlockState> exe(GlobalState globalState) {
            List<BlockState> allStates = globalState.getEntryBlockState().getAllSuccState();
            allStates.remove(globalState.getEntryBlockState());
            return CollectionUtils.where(allStates, new Pred<BlockState>() {
                @Override public Boolean exe(BlockState e) {
                    return !e.isLeaf() && !e.getImmSuccStates().contains(e);
                }
            });
        }
    }

}
