package softmc;

import java.util.Collection;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.Map.Entry;

import korat.utils.func.Pred;
import softmc.logic.Factory;
import softmc.logic.Term;
import softmc.logic.Var;
import softmc.logic.VarModTerm;
import softmc.logic.visitors.TermSubstituter;
import softmc.state.BlockState;
import softmc.state.GlobalState;
import softmc.state.StateTransition;
import softmc.state.SymState;

public class States2ListLogic {

    private GlobalState globalState; 
    private Var stateIdVar;
    private Term nextStateIdVar;
    private Factory factory;
    private Pred<Var> constVarPred;
    
    public States2ListLogic(GlobalState gs) {
        this.globalState = gs; 
        factory = Factory.instance;
        constVarPred = new Pred<Var>() {
            public Boolean exe(Var var) {
                return globalState.isConstVar(var);
            }
        };
        stateIdVar = factory.mkVar(GlobalState.STATE_ID, factory.INT_SORT);
        nextStateIdVar = factory.mkNextNodeVar(stateIdVar);
    }

    /**
     * Returns the constraint that has to hold in the initial (first) state. 
     * The constraint is in terms of the current list node. 
     */
    public Term convertToInitConstr() {
        List<Term> terms = new LinkedList<Term>();
        BlockState entryState = globalState.getEntryBlockState();
        terms.add(factory.mkEq(stateIdVar, factory.mkInt(entryState.getId())));
        for (Entry<String, Term> e : entryState.getSymState().getMappings().entrySet()) {
            String varName = e.getKey();
            Term varValue = e.getValue();
            Var var = globalState.getVarByName(varName);
            terms.add(factory.mkEq(var, varValue));
        }
        if (entryState.isError()) {
            terms.add(factory.mkEndOfList(1));
        } else if (entryState.isLeaf()) {
            terms.add(factory.mkBool(false));
        } else {
            terms.add(factory.mkNotEndOfList(1));
        }
        for (Var in : globalState.getInputParams()) {
            if (in.getVarType().getSortName().equals("int")) {
                terms.add(factory.mkGt(in, factory.mkVar("MIN", "int")));
                terms.add(factory.mkLe(in, factory.mkVar("MAX", "int")));
            }
        }
        return factory.mkAnd(terms.toArray(new Term[0]));
    }
    
    /**
     * Returns a logic expression in terms of variables in the current state (list node)
     * and the next state (also list node).  This logic expression defines valid
     * state transitions.
     */
    public Term convertToTransitionFunction() {
        List<Term> trFn = new LinkedList<Term>();
        for (BlockState bs : globalState.getEntryBlockState().getAllSuccState()) {
            trFn.add(factory.mkImplies(
                    factory.mkEq(stateIdVar, factory.mkInt(bs.getId())), 
                    getTransitionConstraint(bs)));
        }
        Term trFnTerm = factory.mkAnd(trFn.toArray(new Term[0]));
        return TermSubstituter.unwrap(trFnTerm, constVarPred, VarModTerm.class);
    }
    
    /**
     * Returns a logic expression in terms of the current state.  If the current state
     * is the leaf state this expression evaluates to true, otherwise it evaluates
     * to false. 
     */
    @Deprecated
    public Term convertToTermCond() {
        Collection<BlockState> errorStates = globalState.getLeafStates();
        List<Term> termCond = new LinkedList<Term>();
        for (BlockState bs : errorStates) {
            termCond.add(factory.mkEq(stateIdVar, factory.mkInt(bs.getId())));
        }
        return factory.mkOr(termCond.toArray(new Term[0]));
    }
    
    /**
     * Returns a logic expression in terms of the current state.  If the current state
     * is the error state this expression evaluates to true, otherwise it evaluates
     * to false. 
     */
    @Deprecated
    public Term convertToErrorCond() {
        Collection<BlockState> errorStates = globalState.getErrorStates();
        if (errorStates.size() == 0)
            throw new RuntimeException("NO ERROR STATES FOUND!!! CAN'T MODEL CHECK IF NO ERROR STATES");
        List<Term> errorCond = new LinkedList<Term>();
        for (BlockState bs : errorStates) {
            errorCond.add(factory.mkEq(stateIdVar, factory.mkInt(bs.getId())));
        }
        return factory.mkOr(errorCond.toArray(new Term[0]));
    }

    private Term getTransitionConstraint(BlockState bs) {
        List<StateTransition> transitions = bs.getTransitions();
        final int nSucc = transitions.size();

        // no successors => there shouldn't be next state
        if (nSucc == 0) 
            return factory.mkBool(false); 
     
        // single successor
        if (nSucc == 1) 
            return getNextStateAssignments(transitions.get(0));
        
        // multiple successors
        return computeIte(0, transitions);
    }

    private Term computeIte(int idx, List<StateTransition> transitions) {
        StateTransition tr = transitions.get(idx);
        Term nextStateAssignments = getNextStateAssignments(tr);
        if (idx == transitions.size() - 1)
            return nextStateAssignments;
        Term cond = tr.getGuard();
        cond = TermSubstituter.unwrap(cond, constVarPred, VarModTerm.class);
        return factory.mkIte(
                cond, 
                nextStateAssignments, 
                computeIte(idx + 1, transitions));
    }

    private Term getNextStateAssignments(StateTransition transition) {
        BlockState nextState = transition.getDest();
        List<Term> terms = new LinkedList<Term>();
        terms.add(factory.mkEq(nextStateIdVar, factory.mkInt(nextState.getId())));
        Set<String> mutableVarNames = globalState.getMutableVarNamesCopy();
        mutableVarNames.remove(stateIdVar.getVarName());
        SymState effects = SymState.mergeStates(transition.getEffects(), nextState.getSymState());
        for (Entry<String, Term> e : effects.getMappings().entrySet()) {
            String varName = e.getKey();
            mutableVarNames.remove(varName);
            Term varValue = e.getValue();
            Var var = globalState.getVarByName(varName);
            terms.add(factory.mkEq(factory.mkNextNodeVar(var), varValue));
        }
        for (String varName : mutableVarNames) {
            Var var = globalState.getVarByName(varName);
            terms.add(factory.mkEq(factory.mkNextNodeVar(var), var));
        }
        if (nextState.isError())
            terms.add(factory.mkEndOfList(2));
        else 
            if (nextState.isLeaf())
                terms.add(factory.mkBool(false));
            else 
                terms.add(factory.mkNotEndOfList(2));
        return factory.mkAnd(terms.toArray(new Term[0]));
    }
    
}
