package softmc;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.text.MessageFormat;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;

import korat.utils.visitors.IVisitor;
import korat.utils.visitors.VisitorUtils;

import org.apache.commons.io.IOUtils;

import softmc.logic.App;
import softmc.logic.ArrayStore;
import softmc.logic.BinTerm;
import softmc.logic.BoolConst;
import softmc.logic.EndOfList;
import softmc.logic.IntConst;
import softmc.logic.Ite;
import softmc.logic.N_aryTerm;
import softmc.logic.NextNodeVar;
import softmc.logic.NotEndOfList;
import softmc.logic.Sort;
import softmc.logic.Term;
import softmc.logic.UnaryTerm;
import softmc.logic.Var;
import softmc.state.GlobalState;

public class Z3Converter {

    private static final String _NAME = "/*#NAME#*/";
    private static final String _STATE_VARS = "/*#STATE_VARS#*/";
    private static final String _STATE_VAR_TYPES = "/*#STATE_VAR_TYPES#*/";
    private static final String _CONST_DECL = "/*#CONST_DECL#*/";
    private static final String _CONST_DEFN = "/*#CONST_DEFN#*/";
    private static final String _INIT_STATE_CONSTR = "/*#INITIAL_STATE_CONSTR#*/";
    private static final String _RET_TR_FN = "/*#RET_TR_FN#*/";
    
    private final String name; 
    private final GlobalState state;
    private final Term initConstr;
    private final Term trFn;
    private final String templateFileName;
    
    private String templFileContent;  
    private Map<String, String> map; 
    private Map<String, Integer> stateVarToIdx;
    
    public Z3Converter(String name, GlobalState state, Term initConstr, Term trFn, /*Term termCnd, Term errCnd,*/ String templateFileName) {
        this.name = name;
        this.state = state;
        this.initConstr = initConstr;
        this.trFn = trFn;
        this.templateFileName = templateFileName;
    }
    
    public void convert() {
        readTemplateContent();
        produceMap();
        fillOutAndSaveTemplate();
    }
    
    private void fillOutAndSaveTemplate() {
        for (Entry<String, String> e : map.entrySet()) {
            String key = e.getKey().replaceAll("\\*", "\\\\*");
            templFileContent = templFileContent.replaceAll(key, e.getValue());
        }
        String outFileName = String.format("%s/%sChecker.cs", 
                templateFileName.substring(0, templateFileName.lastIndexOf("/")), name);
        try {
            OutputStream fos = new BufferedOutputStream(new FileOutputStream(outFileName));
            IOUtils.write(templFileContent, fos);
            fos.close();
            System.out.println("C# Checker saved to: " + new File(outFileName).getAbsolutePath());
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }
    
    private void produceMap() {
        map = new HashMap<String, String>();
        map.put(_NAME, name);
        String[] stateVarsAndTypes = getStateVarsAndTypes();
        map.put(_STATE_VARS, stateVarsAndTypes[0]);
        map.put(_STATE_VAR_TYPES, stateVarsAndTypes[1]);
        String[] constDeclDefn = getConstDeclDefn();
        map.put(_CONST_DECL, constDeclDefn[0]);
        map.put(_CONST_DEFN, constDeclDefn[1]);
        map.put(_INIT_STATE_CONSTR, getInitStateConstr());
        map.put(_RET_TR_FN, getRetTrFn());
//        map.put(_RET_TERM_COND, getRetTermCond());
//        map.put(_RET_ERR_COND, getRetErrorCond());
    }

//    private String getRetTermCond() {
//        String str = new ToCsharpZ3Converter().getCsharpCode(termCnd, 
//                "GetVarPre(list, tupleProjectionFuncs[{1}])", "%");
//        return String.format("return %s;", str);
//    }
//
//    private String getRetErrorCond() {
//        String str = new ToCsharpZ3Converter().getCsharpCode(errCnd, 
//                "GetVarPre(list, tupleProjectionFuncs[{1}])", "%");
//        return String.format("return %s;", str);
//    }
    
    private String getRetTrFn() {
        String str = new ToCsharpZ3Converter().getCsharpCode(trFn, 
                "GetVarPre(list, tupleProjectionFuncs[{1}])", 
                "GetVarPost(list, tupleProjectionFuncs[{1}])");
        return String.format("return %s;", str);
    }

    private String getInitStateConstr() {
        // nextNodePattern is "%" which would cause an exception if used (so this way we 
        // ensure that that the initial state condition doesn't refer to the next state)
        String str = new ToCsharpZ3Converter().getCsharpCode(initConstr, "GetVarPre(statesList, tupleProjectionFuncs[{1}])", "%");
        return String.format("zStuff.AssertConstr(%s);", str);
    }

    private String[] getConstDeclDefn() {
        StringBuilder constDecl = new StringBuilder();
        StringBuilder constDefn = new StringBuilder();
        for (Var var : state.getVars()) {
            if (!state.isConstVar(var))
                continue;
            String varName = var.getVarName();
            constDecl.append(String.format("protected Term %s;\n", varName));
            constDefn.append(String.format("%s = zStuff.MkConst(\"%s\", %s);\n", varName, varName, getZ3Type(var.getVarType())));
        }
        return new String[] { constDecl.toString(), constDefn.toString() };
    }

    private String[] getStateVarsAndTypes() {
        StringBuilder varNames = new StringBuilder();
        StringBuilder varTypes = new StringBuilder();
        stateVarToIdx = new HashMap<String, Integer>();
        int idx = 0; 
        for (Var var : state.getVars()) {
            if (state.isConstVar(var))
                continue;
            stateVarToIdx.put(var.getVarName(), idx++);
            varNames.append(String.format(", \"%s\"", var.getVarName()));
            varTypes.append(String.format(", %s", getZ3Type(var.getVarType())));
        }
        return new String[] { varNames.substring(2), varTypes.substring(2) };
    }

    private String getZ3Type(Sort varType) {
        // TODO temp, but good enough for now
        return "zStuff." + varType.getSortName().toUpperCase();
    }

    private void readTemplateContent() {
        try {
            InputStream fis = new BufferedInputStream(new FileInputStream(templateFileName));
            templFileContent = IOUtils.toString(fis);
            fis.close();
        } catch (Exception e) {
            throw new RuntimeException("Could not open input template file", e);
        }
    }
    
// =============================================================
    
    class ToCsharpZ3Converter implements IVisitor<String> {
        
        private String currNodePattern;
        private String nextNodePattern;

        public String getCsharpCode(Term t, String currNodeFormat, String nextNodeFormat) {
            this.currNodePattern = currNodeFormat;
            this.nextNodePattern = nextNodeFormat;
            return VisitorUtils.visit(t, this);
        }
        
        public String visitVar(Var var) {
            String varName = var.getVarName();
            Integer idx = stateVarToIdx.get(varName);
            if (idx == null)
                return varName;
            else 
                return MessageFormat.format(currNodePattern, varName, idx);
        }
        
        public String visitBoolConst(BoolConst boolConst) {
            return String.format("zStuff.MkBool(%s)", boolConst.getBoolConst());
        }
        
        public String visitIntConst(IntConst intConst) {
            return String.format("zStuff.MkInt(%s)", intConst.getIntConst());
        }
        
        public String visitUnaryTerm(UnaryTerm term) {
            String innertTermStr = VisitorUtils.visit(term.getTerm(), this);
            return String.format("z3.Mk%s(%s)", term.getClass().getSimpleName(), innertTermStr);
        }
        
        public String visitBinTerm(BinTerm term) {
            String lhs = VisitorUtils.visit(term.getLhs(), this);
            String rhs = VisitorUtils.visit(term.getRhs(), this);
            return String.format("z3.Mk%s(%s, %s)", term.getClass().getSimpleName(), lhs, rhs);
        }
        
        public String visitIte(Ite ite) {
            return String.format("z3.MkIte(%s, %s, %s)", VisitorUtils.visit(ite.getCond(), this), 
                    VisitorUtils.visit(ite.getThenTerm(), this), VisitorUtils.visit(ite.getElseTerm(), this));
        }
        
        public String visitArrayStore(ArrayStore arrayStore) { 
            return String.format("z3.MkArrayStore(%s, %s, %s)", 
                    VisitorUtils.visit(arrayStore.getArray(), this), 
                    VisitorUtils.visit(arrayStore.getIndex(), this),
                    VisitorUtils.visit(arrayStore.getValue(), this));
        }
        
        public String visitApp(App term) {
            StringBuilder sb = new StringBuilder();
            for (Term t : term.getTerms()) {
                String tStr = VisitorUtils.visit(t, this);
                sb.append(", ").append(tStr);
            }
            return String.format("z3.MkApp(%s, new Term[] {%s})", term.getAppName(), sb.substring(2));
        }
        
        public String visitN_aryTerm(N_aryTerm term) {
            StringBuilder sb = new StringBuilder();
            for (Term t : term.getTerms()) {
                String tStr = VisitorUtils.visit(t, this);
                sb.append(", ").append(tStr);
            }
            return String.format("z3.Mk%s(new Term[] {%s})", term.getClass().getSimpleName(), sb.substring(2));
        }
        
        public String visitNextNodeVar(NextNodeVar curr) {
            String varName = curr.getVar().getVarName();
            return MessageFormat.format(nextNodePattern, varName, stateVarToIdx.get(varName));
        }
        
        public String visitEndOfList(EndOfList eol) {
            int hops = eol.getHops();
            if (!(hops >= 1))
                throw new RuntimeException("at least one hop !?!!?!?!?!?!?");
            String z = "z3.MkApp(listTailFunc, list)";
            for (int i = 1; i < hops; i++) 
                z = String.format("z3.MkApp(listTailFunc, %s)", z);
            return String.format("z3.MkApp(isListNilFunc, %s)", z);
        }

        public String visitNotEndOfList(NotEndOfList eol) {
            int hops = eol.getHops();
            if (!(hops >= 1))
                throw new RuntimeException("at least one hop !?!!?!?!?!?!?");
            String z = "z3.MkApp(listTailFunc, list)";
            for (int i = 1; i < hops; i++) 
                z = String.format("z3.MkApp(listTailFunc, %s)", z);
            return String.format("z3.MkApp(isListConsFunc, %s)", z);
        }
        
        public String visitTerm(Term t) {
            System.err.println("All different kinds of terms should have been handled individually");
            System.err.println("Class not handled: " + t.getClass().getName());
            new Throwable().printStackTrace();
            System.exit(2);
            return "";
        }
    }

}
