package softmc.logic;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

public class Factory {
    
    public static final Factory instance = new Factory();

    public final Sort INT_SORT;
    public final Sort BOOL_SORT;
    public final Sort INT_ARRAY_SORT;
    
    private Map<String, Var> allVars = new HashMap<String, Var>();
    private Map<String, Sort> allSorts = new HashMap<String, Sort>();
    
    private Factory() {
        INT_SORT = mkSort("int");
        BOOL_SORT = mkSort("bool");
        INT_ARRAY_SORT = mkSort("int_array");
    }
    
    public Add mkAdd(Term lhs, Term rhs) { return new Add(lhs, rhs); }
    public Term mkAnd(Term... terms) { 
        List<Term> tt = new LinkedList<Term>();
        for (Term t : terms) {
            if (t instanceof BoolConst) {
                BoolConst b = (BoolConst)t; 
                if (!b.getBoolConst())
                    return b;
            } else if (t instanceof And) {
                for (Term subAndTerms : ((And)t).getTerms())
                    tt.add(subAndTerms);
            } else {
                tt.add(t);
            }
        }
        if (tt.size() == 0)
            return mkBool(true);
        else if (tt.size() == 1)
            return tt.get(0);
        else
            return new And(tt.toArray(new Term[0])); 
    }
    public Term mkOr(Term... terms) { 
        List<Term> tt = new LinkedList<Term>();
        for (Term t : terms) {
            if (t instanceof BoolConst) {
                BoolConst b = (BoolConst)t; 
                if (b.getBoolConst())
                    return b;
            } else if (t instanceof Or) {
                for (Term subOrTerms : ((Or)t).getTerms())
                    tt.add(subOrTerms);
            } else {
                tt.add(t);
            }
        }
        if (tt.size() == 0)
            return mkBool(false);
        else if (tt.size() == 1)
            return tt.get(0);
        else
            return new Or(tt.toArray(new Term[0])); 
    }
    public App mkApp(String appName, Sort retSort, Term... args) { return new App(appName, retSort, args); }
    public App mkApp(String appName, String retSortName, Term... args) { return new App(appName, mkSort(retSortName), args); }
    public Div mkDiv(Term lhs, Term rhs) { return new Div(lhs, rhs); }
    public Implies mkImplies(Term lhs, Term rhs) { return new Implies(lhs, rhs); }
    public Ite mkIte(Term cond, Term thenTerm, Term elseTerm) { return new Ite(cond, thenTerm, elseTerm); }
    public IntConst mkInt(int intConst) { return new IntConst(intConst); }
    public BoolConst mkBool(boolean boolConst) { return new BoolConst(boolConst); }
    public Mod mkMod(Term lhs, Term rhs) { return new Mod(lhs, rhs); }
    public Mul mkMul(Term lhs, Term rhs) { return new Mul(lhs, rhs); }
    public NextNodeVar mkNextNodeVar(Var var) { return new NextNodeVar (var); }
    public Sub mkSub(Term lhs, Term rhs) { return new Sub(lhs, rhs); }
    public Eq mkEq(Term lhs, Term rhs) { return new Eq(lhs, rhs); }
    public Term mkNot(Term term) { return new Not(term); }
    public Term mkNeq(Term lhs, Term rhs) { return new Not(new Eq(lhs, rhs)); }
    public Lt mkLt(Term lhs, Term rhs) { return new Lt(lhs, rhs); }
    public Le mkLe(Term lhs, Term rhs) { return new Le(lhs, rhs); }
    public Gt mkGt(Term lhs, Term rhs) { return new Gt(lhs, rhs); }
    public Ge mkGe(Term lhs, Term rhs) { return new Ge(lhs, rhs); }
    
    public Term mkEndOfList(int hops) { return new EndOfList(hops); }
    public Term mkNotEndOfList(int hops) { return new NotEndOfList(hops); }
    
    public Var mkVar(String varName, String varTypeName) { return mkVar(varName, mkSort(varTypeName)); }
    
    public ArraySelect mkArraySelect(Term array, Term index) { return new ArraySelect(array, index); }
    public ArrayStore mkArrayStore(Term array, Term index, Term value) { return new ArrayStore(array, index, value); }
    
    public Var mkVar(String varName, Sort varType) {
        Var var = allVars.get(varName); 
        if (var == null) {
            var = new Var(varName, varType); 
            allVars.put(varName, var);
        }
        return var; 
    }
    
    public Sort mkSort(String sortName) {
        if ("boolean".equals(sortName))
            sortName = "bool";
        // stupid hack: replace trailing "[]" with "_array"
        sortName = sortName.replace("[]", "_array");
        Sort sort = allSorts.get(sortName); 
        if (sort == null) {
            sort = new Sort(sortName); 
            allSorts.put(sortName, sort);
        }
        return sort; 
    }
                 
}
