package softmc.logic.visitors;

import java.util.Map;

import softmc.logic.And;
import softmc.logic.BoolConst;
import softmc.logic.Eq;
import softmc.logic.Ge;
import softmc.logic.Gt;
import softmc.logic.Implies;
import softmc.logic.Le;
import softmc.logic.Lt;
import softmc.logic.Neq;
import softmc.logic.Not;
import softmc.logic.Or;
import softmc.logic.Term;
import softmc.logic.Var;
import korat.utils.visitors.IVisitor;
import korat.utils.visitors.VisitorUtils;

/**
 * To be used to evaluate "condition" terms.
 */
public class CondEvaluator implements IVisitor<Boolean> {
    private final Term term; 
    private final Map<String, Object> values;

    public CondEvaluator(Term term, Map<String, Object> values) {
        this.term = term;
        this.values = values;
    }
    
    public boolean eval() {
        try { 
            return VisitorUtils.visit(term, this);
        } catch (Exception e) {
            throw new RuntimeException("Could not evaluate condition: " + term.toString(), e);
        }
    }

    public Boolean visitBoolConst(BoolConst bc) { return bc.getBoolConst(); }
    public Boolean visitAnd(And and) {
        for (Term t : and.getTerms()) {
            boolean b = evalCond(t);
            if (!b)
                return false;
        }
        return true;
    }
    public Boolean visitOr(Or or) {
        for (Term t : or.getTerms()) {
            boolean b = evalCond(t);
            if (b)
                return true;
        }
        return false;
    }
    public Boolean visitNot(Not not) {
        boolean b = evalCond(not.getTerm());
        return !b;
    }
    public Boolean visitEq(Eq eq) {
        Object lhs = evalExpr(eq.getLhs());
        Object rhs = evalExpr(eq.getRhs());
        return lhs.equals(rhs);
    }
    public Boolean visitNeq(Neq neq) {
        Object lhs = evalExpr(neq.getLhs());
        Object rhs = evalExpr(neq.getRhs());
        return !lhs.equals(rhs);
    }
    public Boolean visitGe(Ge ge) {
        int lhs = evalIntExpr(ge.getLhs());
        int rhs = evalIntExpr(ge.getRhs());
        return lhs >= rhs;
    }
    public Boolean visitGt(Gt gt) {
        int lhs = evalIntExpr(gt.getLhs());
        int rhs = evalIntExpr(gt.getRhs());
        return lhs > rhs;
    }
    public Boolean visitLe(Le le) {
        int lhs = evalIntExpr(le.getLhs());
        int rhs = evalIntExpr(le.getRhs());
        return lhs <= rhs;
    }
    public Boolean visitLt(Lt lt) {
        int lhs = evalIntExpr(lt.getLhs());
        int rhs = evalIntExpr(lt.getRhs());
        return lhs < rhs;
    }
    public Boolean implies(Implies impl) {
        boolean lhs = evalCond(impl.getLhs());
        boolean rhs = evalCond(impl.getRhs());
        return !lhs || rhs;
    }
    public Boolean visitVar(Var var) {
        Object val = values.get(var.getVarName());
        if (val == null) 
            throw new RuntimeException("Unknown value for variable " + var.getVarName());
        return (Boolean) val;
    }
    
    public Object visitTerm(Term t) {
        throw new RuntimeException("Cannot evaluate condition: " + t.toString());
    }
    
    private Boolean evalCond(Term t) {
        return VisitorUtils.visit(t, this);
    }
    
    private Object evalExpr(Term t) {
        return new ExprEvaluator(t, values).eval();
    }
    
    private int evalIntExpr(Term t) {
        return (Integer) evalExpr(t);
    }
}
