package softmc.logic.visitors;

import java.util.Map;

import korat.utils.visitors.IVisitor;
import korat.utils.visitors.VisitorUtils;
import softmc.logic.Add;
import softmc.logic.ArraySelect;
import softmc.logic.ArrayStore;
import softmc.logic.BoolConst;
import softmc.logic.Div;
import softmc.logic.IntConst;
import softmc.logic.Ite;
import softmc.logic.Mod;
import softmc.logic.Mul;
import softmc.logic.Sub;
import softmc.logic.Term;
import softmc.logic.Var;

/**
 * To be used to evaluate "expression" terms.  If the given term to be evaluated
 * contains "non-expression" elements (e.g. conditionals) an exception will be 
 * thrown. 
 */
public class ExprEvaluator implements IVisitor<Object> {

    private final Term term; 
    private final Map<String, Object> values;

    public ExprEvaluator(Term term, Map<String, Object> values) {
        this.term = term;
        this.values = values;
    }
    
    public Object eval() {
        try { 
            return VisitorUtils.visit(term, this);
        } catch (Exception e) {
            e.printStackTrace();
            throw new RuntimeException("Could not evaluate expression: " + term.toString(), e);
        }
    }
    
    public Object visitIntConst(IntConst ic) { return ic.getIntConst(); }
    public Object visitBoolConst(BoolConst bc) { return bc.getBoolConst(); }
    public Object visitAdd(Add add) {
        Integer lhs = (Integer) VisitorUtils.visit(add.getLhs(), this);
        Integer rhs = (Integer) VisitorUtils.visit(add.getRhs(), this);
        return lhs + rhs;
    }
    public Object visitSub(Sub sub) {
        Integer lhs = (Integer) VisitorUtils.visit(sub.getLhs(), this);
        Integer rhs = (Integer) VisitorUtils.visit(sub.getRhs(), this);
        return lhs - rhs;
    }
    public Object visitMul(Mul mul) {
        Integer lhs = (Integer) VisitorUtils.visit(mul.getLhs(), this);
        Integer rhs = (Integer) VisitorUtils.visit(mul.getRhs(), this);
        return lhs * rhs;
    }
    public Object visitDiv(Div div) {
        Integer lhs = (Integer) VisitorUtils.visit(div.getLhs(), this);
        Integer rhs = (Integer) VisitorUtils.visit(div.getRhs(), this);
        return lhs / rhs;
    }
    public Object visitMod(Mod mod) {
        int lhs = (Integer) VisitorUtils.visit(mod.getLhs(), this);
        int rhs = (Integer) VisitorUtils.visit(mod.getRhs(), this);
        return lhs % rhs;
    }
    public Object visitArraySelect(ArraySelect aselect) {
        int[] array = (int[]) VisitorUtils.visit(aselect.getArray(), this);
        int idx = (Integer) VisitorUtils.visit(aselect.getIndex(), this);
        return array[idx];
    }
    public Object visitArrayStore(ArrayStore astore) {
        int[] array = (int[]) VisitorUtils.visit(astore.getArray(), this);
        int[] newArray = new int[array.length];
        System.arraycopy(array, 0, newArray, 0, array.length);
        int idx = (Integer) VisitorUtils.visit(astore.getIndex(), this);
        int value = (Integer) VisitorUtils.visit(astore.getValue(), this);
        newArray[idx] = value; 
        return newArray;
    }
    public Object visitIte(Ite ite) {
        boolean cond = new CondEvaluator(ite.getCond(), values).eval();
        if (cond)
            return VisitorUtils.visit(ite.getThenTerm(), this);
        else
            return VisitorUtils.visit(ite.getElseTerm(), this);
    }
    public Object visitVar(Var var) {
        Object val = values.get(var.getVarName());
        if (val == null) 
            throw new RuntimeException("Unknown value for variable " + var.getVarName());
        return val;
    }
    
    public Object visitTerm(Term t) {
        throw new RuntimeException("Cannot evaluate term: " + t.toString());
    }
}
