package softmc.logic.visitors;

import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import korat.utils.visitors.IVisitor;
import korat.utils.visitors.VisitorUtils;
import softmc.logic.And;
import softmc.logic.BinTerm;
import softmc.logic.BoolConst;
import softmc.logic.Factory;
import softmc.logic.N_aryTerm;
import softmc.logic.Not;
import softmc.logic.Or;
import softmc.logic.Term;
import softmc.logic.UnaryTerm;
import softmc.logic.VarModTerm;

public class Simplifier implements IVisitor<Term> {

    public Term simplify(Term term) {
        return VisitorUtils.visit(term, this);
    }

    public Term visitNot(Not not) {
        Term negatedTerm = not.getTerm();
        negatedTerm = VisitorUtils.visit(negatedTerm, this);
        if (negatedTerm instanceof Not) {
            return ((Not) negatedTerm).getTerm();
        } else {
            not.setTerm(negatedTerm);
            return not;
        }
    }

    public Term visitOr(Or orTerm) {
        List<Term> terms = new LinkedList<Term>();
        for (Term t : orTerm.getTerms()) {
            // remove boolean constants
            if (t instanceof BoolConst) {
                BoolConst b = (BoolConst) t;
                if (b.getBoolConst())
                    return b;
                else
                    continue;
            } else {
                terms.add(VisitorUtils.visit(t, this));
            }
        }
        if (terms.size() == 0)
            return Factory.instance.mkBool(false);

        // search for duplicates
        removeDups(terms);
        if (terms.size() == 1)
            return terms.get(0);

        // search for x, !x
        if (searchOpposites(terms))
            return Factory.instance.mkBool(true);

        // factorize 
        boolean allTermsAnd = true;
        for (Term t : terms) {
            if (!(t instanceof And)) {
                allTermsAnd = false;
                break;
            }
        }
        List<Term> factors = new LinkedList<Term>();
        if (allTermsAnd) {
            And and1 = (And) terms.get(0);
            for (Term t : and1.getTerms()) {
                boolean containsAll = true;
                for (int i = 1; i < terms.size(); i++) {
                    And and2 = (And) terms.get(i);
                    if (!contains(and2.getTerms(), t)) {
                        containsAll = false;
                        break;
                    }
                }
                if (containsAll)
                    factors.add(t);
            }
        }
        Term retTerm; 
        if (factors.isEmpty()) {
            Term[] newTerms = new Term[terms.size()];
            int idx = 0;
            for (Term t : terms) {
                newTerms[idx++] = VisitorUtils.visit(t, this);
            }
            orTerm.setTerms(newTerms);
            retTerm = orTerm;
        } else {
            Term[] newTerms = new Term[factors.size() + 1];
            int idx = 0;
            for (Term t : factors)
                newTerms[idx++] = VisitorUtils.visit(t, this);
            Set<Term> newOrTerms = new HashSet<Term>();
            for (Term t : terms) {
                And andTerm = (And) t;
                Term[] newAndTerms = remove(andTerm.getTerms(), factors);
                newOrTerms.add(Factory.instance.mkAnd(newAndTerms));
            }
            Term mkOr = Factory.instance.mkOr(newOrTerms.toArray(new Term[0]));
            newTerms[idx++] = VisitorUtils.visit(mkOr, this);
            retTerm = Factory.instance.mkAnd(newTerms);
        }
        
        return retTerm;
    }

    public Term visitAnd(And andTerm) {
        List<Term> terms = new LinkedList<Term>();
        for (Term t : andTerm.getTerms()) {
            // remove boolean constants
            if (t instanceof BoolConst) {
                BoolConst b = (BoolConst) t;
                if (!b.getBoolConst())
                    return b;
                else
                    continue;
            } else {
                terms.add(t);
            }
        }
        if (terms.size() == 0)
            return Factory.instance.mkBool(true);

        // search for duplicates
        removeDups(terms);
        if (terms.size() == 1)
            return terms.get(0);

        // search for x, !x
        if (searchOpposites(terms))
            return Factory.instance.mkBool(false);
        
        // distributive law
        boolean allTermsOr = true;
        for (Term t : terms) {
            if (!(t instanceof Or)) {
                allTermsOr = false;
                break;
            }
        }
        List<Term> pullOutTerms = new LinkedList<Term>();
        if (allTermsOr) {
            Or or1 = (Or) terms.get(0);
            for (Term t : or1.getTerms()) {
                boolean containsAll = true;
                for (int i = 1; i < terms.size(); i++) {
                    Or or2 = (Or) terms.get(i);
                    if (!contains(or2.getTerms(), t)) {
                        containsAll = false;
                        break;
                    }
                }
                if (containsAll)
                    pullOutTerms.add(t);
            }
        }
        if (pullOutTerms.isEmpty()) {
            Term[] newTerms = new Term[terms.size()];
            int idx = 0;
            for (Term t : terms) {
                newTerms[idx++] = VisitorUtils.visit(t, this);
            }
            andTerm.setTerms(newTerms);
            return andTerm;
        } else {
            Term[] newTerms = new Term[pullOutTerms.size() + 1];
            int idx = 0;
            for (Term t : pullOutTerms)
                newTerms[idx++] = VisitorUtils.visit(t, this);
            Set<Term> newAndTerms = new HashSet<Term>();
            for (Term t : terms) {
                Or orTerm = (Or) t;
                Term[] newOrTerms = remove(orTerm.getTerms(), pullOutTerms);
                newAndTerms.add(Factory.instance.mkOr(newOrTerms));
            }
            Term mkAnd = Factory.instance.mkAnd(newAndTerms.toArray(new Term[0]));
            newTerms[idx++] = VisitorUtils.visit(mkAnd, this);
            return Factory.instance.mkOr(newTerms);
        }
    }

    private Term[] remove(Term[] terms, List<Term> pullOutTerms) {
        List<Term> newTerms = new LinkedList<Term>();
        for (Term t : terms) {
            if (!pullOutTerms.contains(t))
                newTerms.add(t);
        }
        return newTerms.toArray(new Term[0]);
    }

    private boolean contains(Term[] terms, Term t) {
        for (Term tt : terms) {
            if (tt.equals(t))
                return true;
        }
        return false;
    }

    private void removeDups(List<Term> terms) {
        for (int i = 0; i < terms.size() - 1; i++) {
            Term t1 = terms.get(i);
            for (int j = i + 1; j < terms.size(); j++) {
                if (t1.equals(terms.get(j))) {
                    terms.remove(j);
                    j--;
                }
            }
        }
    }

    private boolean searchOpposites(List<Term> terms) {
        for (int i = 0; i < terms.size() - 1; i++) {
            Term t1 = terms.get(i);
            for (int j = i + 1; j < terms.size(); j++) {
                Term t2 = terms.get(j);
                if (t1.equals(Factory.instance.mkNot(t2)) || t2.equals(Factory.instance.mkNot(t1))) {
                    return true;
                }
            }
        }
        return false;
    }

    public Term visitUnaryTerm(UnaryTerm term) {
        Term innerTerm = VisitorUtils.visit(term.getTerm(), this);
        term.setTerm(innerTerm);
        return term;
    }

    public Term visitBinTerm(BinTerm term) {
        Term lhs = VisitorUtils.visit(term.getLhs(), this);
        Term rhs = VisitorUtils.visit(term.getRhs(), this);
        term.setLhs(lhs);
        term.setRhs(rhs);
        return term;
    }

    public Term visitN_aryTerm(N_aryTerm term) {
        Term[] terms = term.getTerms();
        Term[] newTerms = new Term[terms.length];
        for (int i = 0; i < terms.length; i++) {
            newTerms[i] = VisitorUtils.visit(terms[i], this);
        }
        term.setTerms(newTerms);
        return term;
    }

    public Term visitVarModTerm(VarModTerm term) {
        return term;
    }

    public Term visitTerm(Term term) {
        return term;
    }

}
