package softmc.logic.visitors;

import korat.utils.func.Func1;
import korat.utils.func.Pred;
import korat.utils.reflection.ReflectionUtils;
import korat.utils.visitors.IVisitor;
import korat.utils.visitors.VisitorUtils;
import softmc.logic.BinTerm;
import softmc.logic.N_aryTerm;
import softmc.logic.Term;
import softmc.logic.UnaryTerm;
import softmc.logic.Var;
import softmc.logic.VarModTerm;

/**
 * It doesn't recurse into wrapped variables, i.e. if want to replace "x" with "x+1", 
 * it won't replace "old(x)" with "old(x+1)", because var wrappers can only wrap a single
 * variable, not an arbitrary term.  
 */
public class TermSubstituter implements IVisitor<Term> {
    
    public static final Pred<Var> allVarsPred = new Pred<Var>() {
        @Override public Boolean exe(Var var) {
            return true;
        }
    };

    public static final Pred<Term> isVarPred = new Pred<Term>() {
        @Override public Boolean exe(Term e) {
            if (e instanceof Var)
                return true;
            else
                return false;
        }
    };

    private Func1<Term, Term> substituter;
    
    /**
     * when "substituter" returns null, nothing is substituted
     */
    public Term substitute(Term termToVisit, Func1<Term, Term> substituter) {
        this.substituter = substituter;
        return VisitorUtils.visit(termToVisit, this);
    }

    // ************** utility (helper) methods *****************
    
    public Term substitute(final Term termToVisit, final Pred<Term> matcher, final Func1<Term, Term> substituter) {
        return substitute(termToVisit, new Func1<Term, Term>() {
            @Override
            public Term exe(Term arg) {
                if (matcher.exe(arg))
                    return substituter.exe(arg);
                else
                    return null;
            }
        });
    }
    
    public Term substitute(final Term termToVisit, final Pred<Term> matcher, final Term substituteWith) {
        return substitute(termToVisit, new Func1<Term, Term>() {
            @Override
            public Term exe(Term arg) {
                if (matcher.exe(arg))
                    return substituteWith;
                else
                    return null;
            }
        });
    }

    /**
     * Substitutes all var modifiers (not free variables) with the given another
     * var modifier term (the wrapped variable remains the same).
     */
    @SuppressWarnings("unchecked")
    public static Term rewrappAllVars(final Term termToVisit, final Class<? extends VarModTerm> wrapWith) {
        Pred<Term> myMatcher = new VarModMatcher(allVarsPred, VarModTerm.class);
        Func1<Term, Term> mySubstituter = new VarModCreator(wrapWith);
        return new TermSubstituter().substitute(termToVisit, myMatcher, mySubstituter);
    }

    /**
     * Substitutes all var modifiers (not free variables) that are of the given type  
     * (varModTypeFilter) with the given another var modifier term 
     * (the wrapped variable remains the same).
     */
    @SuppressWarnings("unchecked")
    public static Term rewrap(final Term termToVisit, final Pred<Var> varPred, 
            Class<? extends VarModTerm> varModTypeFilter, 
            Class<? extends VarModTerm> wrapWith) {
        Pred<Term> myMatcher = new VarModMatcher(varPred, varModTypeFilter);
        Func1<Term, Term> mySubstituter = new VarModCreator(wrapWith);
        return new TermSubstituter().substitute(termToVisit, myMatcher, mySubstituter);
    }
    
    public static Term unwrap(final Term termToVisit, final Pred<Var> varPred, 
            Class<? extends VarModTerm> varModTypeFilter) {
        Pred<Term> myMatcher = new VarModMatcher(varPred, varModTypeFilter);
        Func1<Term, Term> mySubstituter = new Func1<Term, Term>() {
            @Override public Term exe(Term arg) {
                return ((VarModTerm) arg).getVar();
            }
        };
        return new TermSubstituter().substitute(termToVisit, myMatcher, mySubstituter);
    }
    
    // ************** visitor methods *****************
    
    public Term visitUnaryTerm(UnaryTerm term) {
        Term subst = substituter.exe(term);
        if (subst != null) {
            return subst;
        } else {
            Term innerTerm = VisitorUtils.visit(term.getTerm(), this);
            term.setTerm(innerTerm);
            return term;
        }
    }

    public Term visitBinTerm(BinTerm term) {
        Term subst = substituter.exe(term);
        if (subst != null) {
            return subst;
        } else {
            Term lhs = VisitorUtils.visit(term.getLhs(), this);
            Term rhs = VisitorUtils.visit(term.getRhs(), this);
            term.setLhs(lhs);
            term.setRhs(rhs);
            return term;
        }
    }

    public Term visitN_aryTerm(N_aryTerm term) {
        Term subst = substituter.exe(term);
        if (subst != null) {
            return subst;
        } else {
            Term[] terms = term.getTerms();
            Term[] newTerms = new Term[terms.length];
            for (int i = 0; i < terms.length; i++) {
                newTerms[i] = VisitorUtils.visit(terms[i], this);
            }
            term.setTerms(newTerms);
            return term;
        }
    }

    public Term visitVarModTerm(VarModTerm term) {
        Term subst = substituter.exe(term);
        if (subst != null) {
            return subst;
        } else {
            return term;
        }
    }

    public Term visitTerm(Term term) {
        Term subst = substituter.exe(term);
        if (subst != null) {
            return subst;
        } else {
            return term;
        }
    }
    
    // ================================================================
    // =================== VarModCreator class ========================
    // ================================================================
    
    static class VarModCreator<E extends VarModTerm> implements Func1<VarModTerm, E> {
        private final Class<E> clz;
        public VarModCreator(Class<E> clz) {
            this.clz = clz;
        }
        @Override
        public E exe(VarModTerm arg) {
            return ReflectionUtils.newInstance(clz, Var.class, arg.getVar());
        } 
    }
    
    // ================================================================
    // =================== VarModMatcher class ========================
    // ================================================================
    
    public static class VarModMatcher implements Pred<Term> {
        private final Pred<Var> varPred; 
        private final Class<? extends VarModTerm> varModTypeFilter;

        public VarModMatcher(final String varName, final Class<? extends VarModTerm> varModTypeFilter) {
            this(new Pred<Var>() {
                @Override
                public Boolean exe(Var e) {
                    return e.getVarName().equals(varName);
                }
            }, varModTypeFilter);
        }
        
        public VarModMatcher(Pred<Var> varPred, Class<? extends VarModTerm> varModTypeFilter) {
            this.varPred = varPred;
            this.varModTypeFilter = varModTypeFilter;
        }

        @Override
        public Boolean exe(Term term) {
            if (!(term instanceof VarModTerm))
                return false;
            VarModTerm varMod = (VarModTerm) term;
            Var var = varMod.getVar();
            if (!varPred.exe(var))
                return false;
            if (!varModTypeFilter.isAssignableFrom(term.getClass()))
                return false;
            return true;
        }

    }
    
    // ================================================================
    // =================== VarMatcher class ========================
    // ================================================================
    
    public static class VarMatcher implements Pred<Term> {
        private final Pred<Var> varPred; 

        public VarMatcher(final String varName) {
            this(new Pred<Var>() {
                @Override
                public Boolean exe(Var e) {
                    if (varName == null || varName.length() == 0)
                        return true;
                    return e.getVarName().equals(varName);
                }
            });
        }
        
        public VarMatcher(Pred<Var> varPred) {
            this.varPred = varPred;
        }

        @Override
        public Boolean exe(Term term) {
            if (!(term instanceof Var))
                return false;
            Var var = (Var) term;
            return varPred.exe(var);
        }
    }

}
