package softmc.state;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import korat.utils.collection.CollectionUtils;
import korat.utils.func.Func1;
import softmc.logic.Factory;
import softmc.logic.Term;
import softmc.logic.visitors.Simplifier;

public class BlockState {

    private static int autoId = 0; 
    
    private final int id = autoId++;
    
    private SymState symState = new SymState();    
    private List<StateTransition> transitions = new ArrayList<StateTransition>(10);
    
    private boolean isError = false;

    private Factory factory = Factory.instance;
    
    public int getId() { return id; }    
    public List<StateTransition> getTransitions() { return transitions; }
    public StateTransition getTransition(int i) { return transitions.get(i); }
    
    /**
     * Use with extreme caution
     */
    public static void resetAudoId() { autoId = 0; }
    
    /**
     * Returns only the immediate successor states.
     */
    public List<BlockState> getImmSuccStates() {
        return CollectionUtils.select(transitions, new Func1<StateTransition, BlockState>() {
            @Override
            public BlockState exe(StateTransition arg) {
                return arg.getDest();
            }
        });
    }
    
    /**
     * Traverses the wholes state graph (in DFS manner) and returns 
     * all reachable successor states INCLUDING this one.  
     */
    public List<BlockState> getAllSuccState() {
        LinkedList<BlockState> allSuccStates = new LinkedList<BlockState>(); 
        LinkedList<BlockState> workList = new LinkedList<BlockState>();
        workList.add(this);
        Set<BlockState> visited = new HashSet<BlockState>();
        while (!workList.isEmpty()) {
            BlockState bs = workList.removeFirst();
            if (!visited.add(bs))
                continue;
            allSuccStates.add(bs);
            for (BlockState succ : bs.getImmSuccStates()) {
                if (succ != null)
                    workList.push(succ);
            }
        }
        return allSuccStates;
    }

    /**
     * Sets two-way branch condition for this state. 
     * 
     * NOTE: make sure to eventually set the two successor states by calling set2waySuccStates.
     */
    public void set2wayBranch(Term cond, BlockState thenState, BlockState elseState) {
        transitions.clear();
        transitions.add(new StateTransition(this, thenState, cond));
        transitions.add(new StateTransition(this, thenState, factory.mkNot(cond)));
    }
    
    public void setNoSucc() {
        transitions.clear();
    }
    
    public void setSingleSuccState(BlockState succState) {
        transitions.clear();
        transitions.add(new StateTransition(this, succState, factory.mkBool(true)));
    }
    
    public void splitSucc(int succBlockIdx, Term cond, BlockState thenState, BlockState elseState) {
        StateTransition oldTr = transitions.get(succBlockIdx);
        StateTransition t1 = new StateTransition(this, thenState, factory.mkAnd(oldTr.getGuard(), cond)); 
        StateTransition t2 = new StateTransition(this, elseState, factory.mkAnd(oldTr.getGuard(), factory.mkNot(cond)));
        transitions.set(succBlockIdx, t1);
        transitions.add(succBlockIdx + 1, t2);
    }
    
    /**
     * @return -1 if the given state is not in the succ list
     */
    public int findSuccStateIdx(BlockState succToFind) {
        int idx = 0; 
        for (StateTransition tr : transitions) {
            if (tr.getDest() == succToFind)
                return idx;
            idx++;                
        }
        return -1;
    }

    public SymState getSymState() {
        return symState; 
    }
    
    public Term getVarSymVal(String varName) {
        return symState.getSymValueForVariable(varName);
    }
    
    public void setVarSymVal(String varName, Term symVal) {
        symState.setSymValueForVariable(varName, symVal);
    }

    public boolean isError() {
        return isError;
    }

    public boolean isLeaf() {
        return transitions.isEmpty();
    }

    public void setError(boolean isError) {
        this.isError = isError;
    }


    public void simplifyAllTerms() {
        Simplifier simplifier = new Simplifier();
        for (StateTransition tr : transitions) {
            tr.setGuard(simplifier.simplify(tr.getGuard()));
        }
        symState.simplifyTerms(simplifier);
    }

    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder("\n");
        if (isError)
            sb.append("<ERROR>\n");
        sb.append("id: " + id + "\n");
        sb.append(symState);
        return sb.substring(1);
    }

    public void update2wayBranchDests(BlockState thenState, BlockState elseState) {
        if (thenState == null && elseState == null) 
            setNoSucc();
        else if (thenState == null)
            setSingleSuccState(elseState);
        else if (elseState == null)
            setSingleSuccState(thenState);
        else {
            /* transitions must be created at some earlier point */
            transitions.get(0).setDest(thenState);
            transitions.get(1).setDest(elseState);
        }
    }
    
    /**
     * @requires trans.getDest() == trans.getSource()
     */
    public void unfoldTransition(StateTransition trans) {
        int idx = transitions.indexOf(trans);
        if (idx == -1)
            throw new RuntimeException("Transition not found in this state");
        if (trans.getSource() != trans.getDest())
            throw new RuntimeException("Cannot unfold non-looping transition");
        StateTransition newTrans = new StateTransition(trans.getSource(), trans.getDest());
        SymState mergedEffects = SymState.mergeStates(trans.getEffects(), trans.getDest().getSymState());
        SymState newEffects = SymState.mergeStates(mergedEffects, trans.getEffects());
        Term newGuard = factory.mkAnd(trans.getGuard(), SymState.updateExpr(trans.getGuard().clone(), mergedEffects));
        newTrans.setGuard(newGuard);
        newTrans.appendEffects(newEffects);
        transitions.add(idx, newTrans);
    }
    
//    /**
//     * @requires trans.getDest() == trans.getSource()
//     */
//    public void unfoldTransition(StateTransition trans) {
//        splitTransition(trans, new LinkedList<StateTransition>(transitions));
//    }

    public boolean splitTransition(StateTransition tr, List<StateTransition> dests) {
        int idx = transitions.indexOf(tr);
        if (idx == -1)
            throw new RuntimeException("State not found");
        return splitTransition(idx, dests);
    }
    
    public boolean splitTransition(int i, List<StateTransition> dests) {
        if (dests.isEmpty())
            return false;
        StateTransition trI = transitions.get(i);
        Term condI = trI.getGuard();
        for (int j = 0; j < dests.size(); j++) {
            StateTransition succSuccJ = dests.get(j);
            Term updatedSuccJGuard = SymState.updateExpr(succSuccJ.getGuard().clone(), trI.getEffects());
            StateTransition newTrJ = new StateTransition(this, succSuccJ.getDest(), 
                    factory.mkAnd(condI, updatedSuccJGuard));
            newTrJ.appendEffects(trI.getEffects());
            newTrJ.appendEffects(succSuccJ.getEffects());
            if (j == 0)
                transitions.set(i, newTrJ);
            else
                transitions.add(i, newTrJ);
        }
        return true;
    }
    
}
