package softmc.state;

import java.util.Collection;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import korat.utils.collection.CollectionUtils;
import korat.utils.func.Func1;
import korat.utils.func.Pred;
import softmc.logic.Factory;
import softmc.logic.Var;

public class GlobalState {

    public static final String STATE_ID = "id";
    
    private final Var stateIdVar;
    private BlockState entryBlockState;
    private LinkedHashMap<String, Var> allVars = new LinkedHashMap<String, Var>();
    private Set<String> mutableVarNames;
    private List<Var> inputParams = new LinkedList<Var>();

    public GlobalState() {
        stateIdVar = Factory.instance.mkVar(STATE_ID, Factory.instance.INT_SORT);
        allVars.put(STATE_ID, stateIdVar);
    }

    public void addVar(Var var) { allVars.put(var.getVarName(), var); }
    public Collection<Var> getVars() { return allVars.values(); }
    
    public Set<String> getMutableVarNamesCopy() {
        return new LinkedHashSet<String>(mutableVarNames);
    }
    
    public boolean isConstVar(Var var) { 
        return allVars.containsKey(var.getVarName()) && !mutableVarNames.contains(var.getVarName()); 
    }
    
    public BlockState getEntryBlockState() {
        return entryBlockState;
    }

    public void setEntryBlockState(BlockState entryBlockState) {
        this.entryBlockState = entryBlockState;
    }

    public Collection<BlockState> getErrorStates() {
        return CollectionUtils.where(entryBlockState.getAllSuccState(), 
                new Pred<BlockState>() {
                    public Boolean exe(BlockState e) {
                        return e.isError();
                    }
                });
    }
    
    public Collection<BlockState> getLeafStates() {
        return CollectionUtils.where(entryBlockState.getAllSuccState(), 
                new Pred<BlockState>() {
                    public Boolean exe(BlockState e) {
                        return e.isLeaf();
                    }
                });
    }

    public BlockState findStateById(final int id) {
        List<BlockState> states = CollectionUtils.where(entryBlockState.getAllSuccState(), new Pred<BlockState>() {
            @Override public Boolean exe(BlockState e) {
                return e.getId() == id;
            }
        });
        if (states.isEmpty())
            return null;
        return states.get(0);
    }
    
    public List<Var> getInputParams() {
        return inputParams;
    }
    
    public void addInputParam(Var var) {
        inputParams.add(var);
    }
    
    /**
     * Mutable variables are those that get assigned in any of the contained Block States
     * 
     * @requires "entryBlockState != null"
     */
    public void computeMutableVars() {
        if (entryBlockState == null)
            return;
        mutableVarNames = new LinkedHashSet<String>();
        mutableVarNames.add(stateIdVar.getVarName());
        for (BlockState bs : entryBlockState.getAllSuccState()) {
            mutableVarNames.addAll(bs.getSymState().getVarNames());
        }
    }
    
    public Var getVarByName(String varName) {
        return allVars.get(varName);
    }
    
    public Collection<StateTransition> getAllTransitions() {
        List<StateTransition> transitions = new LinkedList<StateTransition>();
        for (BlockState bs : entryBlockState.getAllSuccState()) {
            transitions.addAll(bs.getTransitions());
        }
        return transitions;
    }
    
    /**
     * Eliminates the given state by updating outgoing transitions of all of 
     * its predecessor states. Updated transitions will have the symbolic state
     * of the given block state appended in their "effect"s. 
     */
    public void eliminateState(final BlockState bs) {
        if (bs == entryBlockState)
            throw new RuntimeException("Cannot eliminate entry state");
        if (bs.isLeaf())
            throw new RuntimeException("Cannot eliminate leaf state");
        if (!entryBlockState.getAllSuccState().contains(bs))
            throw new RuntimeException("The given state is not in this state graph");
        List<StateTransition> outTrans = bs.getTransitions();
        List<BlockState> succStates = CollectionUtils.select(outTrans, new Func1<StateTransition, BlockState>() {
            @Override public BlockState exe(StateTransition arg) {
                return arg.getDest();
            }
        });
        if (succStates.contains(bs))
            throw new RuntimeException("Cannot eliminate state with self loop");
        
        Collection<StateTransition> trans = getAllTransitions();
        Collection<StateTransition> inTrans = CollectionUtils.where(trans, new Pred<StateTransition>() {
            @Override public Boolean exe(StateTransition e) {
                return e.getDest() == bs;
            }
        });
        
        for (StateTransition inTr : inTrans) {
            inTr.appendEffects(bs.getSymState());
            inTr.getSource().splitTransition(inTr, outTrans);
        }
        
    }

    public String toDot() {
        StringBuilder sb = new StringBuilder();
        Set<BlockState> visited = new HashSet<BlockState>();
        List<StateTransition> visitedTransitions = new LinkedList<StateTransition>();
        sb.append("digraph SymStates {\n");
        visitBlockState(entryBlockState, sb, visited, visitedTransitions);
        StringBuilder sss = new StringBuilder("<<table BORDER=\"0pt\" CELLBORDER=\"0pt\" CELLPADDING=\"3pt\" CELLSPACING=\"0pt\">");
        for (int i = 0; i < visitedTransitions.size(); i++) {
            StateTransition tr = visitedTransitions.get(i);
            sss.append(String.format("<tr><td align=\"right\">c%s:</td><td align=\"left\">%s</td></tr>", 
                    i, tr.toString().replaceAll("<", "&lt;").replaceAll(">", "&gt;").replaceAll("\\n", "<br/>")));
        }
        sss.append("</table>>");        
        sb.append(String.format("  %s [shape=plaintext, label=%s];\n", "legend", sss.toString()));
        sb.append("}\n");
        return sb.toString();
    }

    private static void visitBlockState(BlockState blockState, StringBuilder sb, Set<BlockState> visited, List<StateTransition> visitedConditions) {
        if (!visited.add(blockState))
            return;
        sb.append(String.format("  %s [shape=box, label=<<table BORDER=\"0pt\" CELLBORDER=\"0pt\" CELLPADDING=\"3pt\" CELLSPACING=\"0pt\"><tr><td align=\"right\">%s</td></tr></table>>];\n", getBlockId(blockState), 
                blockState.toString().replaceAll("<", "&lt;").replaceAll(">", "&gt;").replaceAll("\\n", "</td></tr><tr><td align=\"right\">").replaceAll(":", ":</td><td align=\"left\">")));
        for (StateTransition tr : blockState.getTransitions()) {
            BlockState succ = tr.getDest();
            visitedConditions.add(tr);
            sb.append(String.format("  %s -> %s [label=\"c%s\"];\n", getBlockId(blockState), getBlockId(succ), visitedConditions.size() - 1));
            visitBlockState(succ, sb, visited, visitedConditions);
        }
    }

    private static String getBlockId(BlockState bb) {
        return String.format("bs_%s", bb.getId());
    }

}
