package softmc.state;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;

import korat.utils.func.Func1;
import softmc.logic.Term;
import softmc.logic.Var;
import softmc.logic.visitors.Simplifier;
import softmc.logic.visitors.TermSubstituter;

public class SymState implements Cloneable {

    private Map<String, Term> symValues = new HashMap<String, Term>();

    public boolean isEmpty() { return symValues.isEmpty(); }

    public Term getSymValueForVariable(String varName) {
        return symValues.get(varName);
    }

    public void setSymValueForVariable(String varName, Term symVal) {
        symValues.put(varName, symVal);
    }

    public void simplifyTerms(Simplifier simplifier) {
        for (Entry<String, Term> e : symValues.entrySet()) {
            symValues.put(e.getKey(), simplifier.simplify(e.getValue()));
        }
    }

    public Collection<? extends String> getVarNames() {
        return symValues.keySet();
    }

    public Map<String, Term> getMappings() {
        return symValues;
    }
    
    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        for (Entry<String, Term> e : symValues.entrySet()) {
            sb.append(String.format("\n%s: %s", e.getKey(), e.getValue()));
        }
        if (sb.length() == 0)
            return "";
        return sb.substring(1);
    }

    /**
     * Returns a new SymState that is the result of merging state1 and state2 
     * (which both stay unmodified)
     */
    public static SymState mergeStates(SymState state1, SymState state2) {
        SymState ret = new SymState();
        appendTo(ret, state1);
        appendTo(ret, state2);
        return ret;
    }
    
    /**
     * Applies beta to alpha.
     */
    private static void appendTo(SymState alpha1, final SymState alpha2) {
        // clone alpha2 symbolic map
        Map<String, Term> alpha2Map = new HashMap<String, Term>();
        for (Entry<String, Term> e : alpha2.symValues.entrySet()) {
            alpha2Map.put(e.getKey(), e.getValue().clone());
        }
        
        updateAction(alpha2Map, alpha1.symValues);
        
        // finally, just add (or overwrite if already exists) these new, updated 
        // terms to lhs
        for (Entry<String, Term> e : alpha2Map.entrySet()) {
            alpha1.setSymValueForVariable(e.getKey(), e.getValue());
        }
    }

    /**
     * updates all assignments in alpha with respect to beta
     */
    public static void updateAction(SymState alpha, SymState beta) {
        updateAction(alpha.getMappings(), beta.getMappings());
    }
    
    private static void updateAction(Map<String, Term> alpha, Map<String, Term> beta) {
        for (Entry<String, Term> e : alpha.entrySet()) {
            e.setValue(updateExpr(e.getValue(), beta));
        }
    }

    /**
     * updates all vars in expr with respect to beta. 
     */
    public static Term updateExpr(Term expr, final SymState beta) {
        return updateExpr(expr, beta.getMappings());
    }
    
    private static Term updateExpr(Term expr, final Map<String, Term> beta) {
        TermSubstituter subst = new TermSubstituter();
        Func1<Term, Term> substituter = new Func1<Term, Term>() {
            @Override public Term exe(Term term) {
                if (!(term instanceof Var))
                    return null;
                Var var = (Var) term;
                Term substExpr = beta.get(var.getVarName());
                if (substExpr == null)
                    return null;
                else
                    return substExpr.clone();
            }
        };
        return subst.substitute(expr, substituter);
    }
    
}
