package softmc.tools;

import java.io.PrintStream;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import korat.utils.collection.CollectionUtils;
import korat.utils.collection.Pair;
import korat.utils.func.Pred;
import softmc.ModelChecker;
import softmc.Utils;
import softmc.examples.Examples;
import softmc.logic.Term;
import softmc.logic.visitors.CondEvaluator;
import softmc.logic.visitors.ExprEvaluator;
import softmc.state.BlockState;
import softmc.state.GlobalState;
import softmc.state.StateTransition;
import softmc.state.SymState;

public class Interpreter {
    
    private static final String ID = "#id#";
    private final GlobalState state; 
    private final Map<String, Object> vals;
    
    public Interpreter(GlobalState state, Map<String, Object> initValuations) {
        this.state = state;
        this.vals = initValuations;
    } 

    public ExePath go() {
        ExePath path = new ExePath();
        BlockState currState = state.getEntryBlockState();
        StateTransition prevTrans = null;
        while (true) {
            // update state
            updateState(currState.getSymState());
            vals.put(ID, currState.getId());
            addTraceElem(path, prevTrans);
            
            // figure out next transition
            if (currState.isLeaf()) {
                path.setInErrorState(currState.isError());
                break;
            }
            
            List<StateTransition> possibleTransitions = CollectionUtils.where(currState.getTransitions(), new Pred<StateTransition>() {
                @Override public Boolean exe(StateTransition e) {
                    return new CondEvaluator(e.getGuard(), vals).eval();
                }
            });
            
            // take the transition
            if (possibleTransitions.isEmpty()) {
                path.setError("No available transitions from here");
                break;
            }
            if (possibleTransitions.size() > 1) {
                StringBuilder sb = new StringBuilder("More than one possible transition from here:");
                for (StateTransition t : possibleTransitions)
                    sb.append("  " + t).append("\n");
                path.setError(sb.toString());
                break;
            }
            prevTrans = possibleTransitions.get(0);
            updateState(prevTrans.getEffects());
            currState = prevTrans.getDest();
        }
        return path;
    }

    private void addTraceElem(ExePath path, StateTransition prevTrans) {
        Map<String, Object> values = new HashMap<String, Object>();
        for (Entry<String, Object> e : vals.entrySet()) {
            values.put(e.getKey(), e.getValue());
        }
        path.add(new Pair<StateTransition, Map<String,Object>>(prevTrans, values));
    }
    
    private Map<String, Object> updateState(SymState symState) {
        // compute valuations for this state
        Map<String, Object> currVals = new HashMap<String, Object>();
        for (Entry<String, Term> e : symState.getMappings().entrySet()) {
            currVals.put(e.getKey(), eval(e.getValue()));
        }
        // update global state
        for (Entry<String, Object> e : currVals.entrySet()) {
            vals.put(e.getKey(), e.getValue());
        }
        return currVals;
    }

    private Object eval(Term term) {
        return new ExprEvaluator(term, vals).eval();
    }

    public static void main(String[] args) {
        String className = Examples.class.getName();
        String methodName = "bblSort";
        System.out.println(String.format("Checking method %s::%s", className, methodName));
        ModelChecker mc = new ModelChecker();
        try {
            GlobalState gs = mc.convertToState(className, methodName);
            Utils.dot2Pdf(gs.toDot(), "x.pdf");
            Map<String, Object> vals = new HashMap<String, Object>();
            vals.put("N0", 3);
            vals.put("a1", new int[] { 6, 7, 4 });
            new Interpreter(gs, vals).go().print();
        } catch (Exception e) {
            e.printStackTrace();
        }
    }
    
    // ==================================================================
    // ---------------------- ExePath class -----------------------------
    // ==================================================================
    
    public static class ExePath implements Iterable<Pair<StateTransition, Map<String, Object>>> {
        private LinkedList<Pair<StateTransition, Map<String, Object>>> path;
        private String errorMessage = null;
        private boolean inErrorState = false;
        
        public ExePath() {
            path = new LinkedList<Pair<StateTransition,Map<String,Object>>>();
        }
        
        public String getErrorMessage() {
            return errorMessage;
        }

        public void setErrorMessage(String errorMessage) {
            this.errorMessage = errorMessage;
        }

        public boolean isInErrorState() {
            return inErrorState;
        }

        public void setInErrorState(boolean reachedErrorState ) {
            this.inErrorState = reachedErrorState; 
        }

        public void setError(String msg) { errorMessage = msg; }

        public boolean add(Pair<StateTransition, Map<String, Object>> e) {
            return path.add(e);
        }

        public void clear() {
            path.clear();
        }

        public Pair<StateTransition, Map<String, Object>> get(int index) {
            return path.get(index);
        }

        public Iterator<Pair<StateTransition, Map<String, Object>>> iterator() {
            return path.iterator();
        }

        public int size() {
            return path.size();
        }
        
        public Pair<StateTransition, Map<String, Object>> getLast() {
            return path.getLast();
        }
        
        public void print() { print(System.out); }
        
        public void print(PrintStream out) {
            for (Pair<StateTransition, Map<String, Object>> pathElem : path) {
                if (pathElem.getKey() != null) {
                    out.println("\n --> Taking transition: " + pathElem.getKey());
                    out.println("=======================================");
                }
                printValuations(out, pathElem.getValue());
            }
            if (errorMessage != null) {
                out.println("ERROR: ");
                out.println(errorMessage);
                throw new RuntimeException(errorMessage);
            } else {
                if (inErrorState)
                    out.println("************ REACHED AN ERROR STATE ************");
                else
                    out.println("~~~~~~~~~~~ END OF EXECUTION ~~~~~~~~~~~");
            }
        }
        
        private void printValuations(PrintStream out, Map<String, Object> vals) {
            out.println(String.format(" %s: %s", ID, vals.get(ID)));
            for (Entry<String, Object> e : vals.entrySet()) {
                if (e.getKey().equals(ID))
                    continue;
                Object value = e.getValue();
                if (value.getClass().isArray())
                    value = Arrays.toString((int[]) value);
                out.println(String.format(" %s: %s", e.getKey(), value));
            }
        }
    }
}
